/* Unix-Unix Simple Heirarchial Reordering Decoder version 3.2
   Written by Mark Neidengard, mneideng@cosmos.acs.calpoly.edu
   uuconvert routines by permission of weigley@sun.acsu.buffalo.edu
   July, 1993.  Permission to distribute and use this code
   freely as long as source remains unchanged.  Please submit
   all suggestions to me via email.   */

#define TRUE    1
#define FALSE   0
#define DEC(c)  (((c) - ' ') & 077)
#define LENGTH  150

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
FILE *out;
#include "source/structs.h"
#include "source/initargs.c"
#include "source/handleargs.c"
#include "source/messages.c"
#include "source/subject.c" 
#include "source/correlate.c"
#include "source/decode.c"

 
int main (argc, argv)
int argc;
char *argv[];
{
 
   /*  all internal variable definitions go here */
struct switches main_switch;
int arg_count;    /* what argument number we're on */
int subject_no=0; /* the last subject in the list */
int read_me=1;
int proper_length;
int line_size=sizeof (struct data_line); /* save calculation later */
int out_count=0; /* for output subject counting */
FILE *infile, *outfile, *fopen (); /* the current source of input */ 
char buffer[91];
struct subject_line subject_index[1024];
struct data_line *current, *first;

 
   initargs (&main_switch); /* set all switches off */
   current=(struct data_line *) calloc (1,line_size);
   first=current; /* make sure we keep track of the start of the data */
 
   if (argc < 2 ) {
     help_screen();
         exit (1);
   }
   if ((arg_count=handleargs(argc,argv,&main_switch))==argc)
      if (!main_switch.pipeline) {
         no_files(argv[0]);
         exit (2);
      }

/* now the fun begins...we start reading data from the inputs */
   infile=0; 
   if (main_switch.phase_info) printf ("Getting subjects...\n");
   do {
      if (infile) fclose(infile);
      if (arg_count==argc) infile=stdin;
      else 
         if ((infile=fopen (argv[arg_count],"r"))==NULL) {
            file_error(argv[0],argv[arg_count]);
            exit (3);
         }
      while (fgets (buffer,90,infile)) {
         if (!strncmp (buffer,"Subject:",7)) {
            read_me=subject_handle (buffer,subject_index,subject_no,&main_switch);
            if (read_me) {
               subject_index[subject_no].index=current;
               if (main_switch.input_sub) printf ("%s",buffer);
               subject_no+=read_me;
               current->is_subject=1;
            }
         }
         if (read_me) {
            strcpy (current->string,buffer);
            current->next=(struct data_line *) calloc (1,line_size);
            current=current->next; 
         }
      }
   }
   while (++arg_count < argc);

   if (!subject_no) {
      printf ("%s: No subjects present\n",argv[0]);
      exit (4);
   }

/* having read the data, we sort the subjects with quicksort */
       if (main_switch.phase_info) printf ("Sorting...\n");
   quicksort (subject_index,0,subject_no-1);
/* and then, we correlate them.. */
   if (main_switch.correlate) {
      if (main_switch.phase_info) printf ("Correlating...\n");
      correlate (subject_index,subject_no);
   }
   if (main_switch.sorted_sub) {
      while (out_count < subject_no) {
         if (subject_index[out_count].correlated) printf ("---> ");
      printf ("%s",subject_index[out_count].string);
      ++out_count;
      }
   }

/* and finally, we output using the subject_index as a guide */

   out_count=0;
   if (main_switch.phase_info) printf ("Output phase...\n");

   if (main_switch.correlate) {
      if (!(*(main_switch.uncorrelated_out)))
         outfile=stdout;
      else {
        outfile=fopen (main_switch.uncorrelated_out,"a");
        if (main_switch.phase_info) {
           printf ("Unmatched parts to ");
           if (!(*(main_switch.uncorrelated_out)))
              printf ("stdout\n");
           else printf ("%s\n",main_switch.uncorrelated_out);
        }
      }
      current=first;
      while (out_count<subject_no) {
         if (!subject_index[out_count].correlated) {
            ++out_count;
            continue;
         }
         fprintf (outfile,"%s",subject_index[out_count].index->string);
         if (subject_index[out_count].index->next)

subject_index[out_count].index=subject_index[out_count].index->next;
         if (subject_index[out_count].index->is_subject || !subject_index[out_count].index->next)
            ++out_count;
      }
   }

   out_count=0;
   if (main_switch.no_decode) {
      if (!(*(main_switch.undecoded_out)))
         outfile=stdout;
      else {
        outfile=fopen (main_switch.undecoded_out,"a");
        if (main_switch.phase_info) {
           printf ("Undecoded uue's to ");
           if (!(*(main_switch.uncorrelated_out)))
              printf ("stdout\n");
           else printf ("%s\n",main_switch.undecoded_out);
        }
        }
      current=first;
      while (out_count<subject_no) {
         if (subject_index[out_count].correlated) {
            ++out_count;
            continue;
         }
         fprintf (outfile,"%s",subject_index[out_count].index->string);
         if (subject_index[out_count].index->next)
         subject_index[out_count].index=subject_index[out_count].index->next;
         if (subject_index[out_count].index->is_subject || !subject_index[out_count].index->next)
            ++out_count;
      }
   }
   else {
  char *s2, *s1, *s0, *tmp_s;
  int length;
  static char s[3 * (LENGTH + 1)];
  static int echo_on = FALSE, started = FALSE, just_finished = FALSE;
  static int line_length = 0, lines_to_go = 0, firstline=FALSE;
 
  s0 = s;
  s1 = s0 + (LENGTH + 1);
  s2 = s1 + (LENGTH + 1);
 
  s0[0] = s1[0] = s2[0] = '\0';  /* Clear strings */
 
  while (out_count<subject_no && subject_index[out_count].index->string)
   { 
   if (subject_index[out_count].correlated) {
      ++out_count;
      continue;
   }
   strcpy (s0,subject_index[out_count].index->string);
 

subject_index[out_count].index=subject_index[out_count].index->next;
   if (subject_index[out_count].index->is_subject || !subject_index[out_count].index->next)
      ++out_count;
 
 
 
    s0[LENGTH] = '\0';  /* Make sure string is terminated */
 
    if (just_finished)
      just_finished = FALSE; /* yea I don't why either */
 
    if (!started)
      {
        if (strncmp(s0, "begin ", 6) == 0)
          {
            int mode;
            char dest[128];
            firstline=TRUE;
            started = echo_on = TRUE, line_length = 0, lines_to_go = 0;
            sscanf(s0, "begin %o %s", &mode, dest);
            out = fopen(dest, "r");
            if (out != NULL) {
               strcat (dest,".new");
            fclose (out);
            }
            out = fopen(dest, "w");
            if (main_switch.files_made) printf ("%s\n",dest);
          }
      }
    else  /* started */
    {
      length = strlen(s0);
      if (line_length == 0)
        line_length = length;
 
      if (echo_on)
      {
        lines_to_go = 0;
        if (s0[0] != 'M' || length != line_length)
        {
          echo_on = FALSE;
          lines_to_go = 2;  /* Lines to go before 'end' is expected */
          if (s0[0] == ' ' || s0[0] == '`')
            lines_to_go = 1;
        }
      }
      else  /* !echo_on */
        if (s0[0] == 'M' && length == line_length)
          echo_on = TRUE;
        else if (lines_to_go > 0)
          if (lines_to_go == 2)
            if (s0[0] == ' ' || s0[0] == '`')
              lines_to_go = 1;
            else
              lines_to_go = 0;  /* Unexpected line, so break off */
          else if (lines_to_go == 1)
            if (strcmp(s0, "end\n") == 0)
            {
           doaline(s2);
              doaline(s1);
              lines_to_go = 0;  /* Done. Break off */
              just_finished = TRUE;
              started = FALSE;
            }
            else
              lines_to_go = 0;  /* Unexpected line, so break off */
    }
    
    if (echo_on&&(!firstline))
      doaline(s0), s0[0] = '\0';
    else if (firstline)
      firstline=FALSE;
    
    tmp_s = s2, s2 = s1, s1 = s0, s0 = tmp_s;
   }
  }
        exit (0);
}
